#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2021
# File   : weapons.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 23.02.2021

import collections
from dataclasses import dataclass, field
from enum import IntEnum, auto

import dice

from ..utils import require_kwargs, reverse_number, d10, d100, Nd10
from .character import Characteristic
from .items import Craftsmanship, InstanceMixin, Item


class WeaponClass(IntEnum):
    Pistol = auto()
    Basic = auto()
    Heavy = auto()
    Thrown = auto()
    Melee = auto()


class DamageType(IntEnum):
    Energy = auto()
    Impact = auto()
    Rending = auto()
    Explosive = auto()


class WeaponType(IntEnum):
    Las = auto()
    SP = auto()
    Bolt = auto()
    Melta = auto()
    Plasma = auto()
    Flame = auto()
    Primitive = auto()
    Launcher = auto()
    Grenade = auto()
    Missile = auto()
    Exotic = auto()
    Xenos = auto()
    Chain = auto()
    Shock = auto()
    Power = auto()
    Force = auto()
    Ork = auto()


class ShotType(IntEnum):
    Single = auto()
    Semi = auto()
    Full = auto()


@require_kwargs
@dataclass(frozen=True)
class PlayerWeapon(Item):

    weapon_class: WeaponClass
    weapon_type: WeaponType
    weapon_range: int
    rof: tuple
    damage_roll: int
    damage_bonus: int
    damage_type: DamageType
    pen: int
    clip: int
    reload_time: float
    special: tuple = field(default_factory=tuple)
    reference: str = ''
    extra: str = ''

    @property
    def pretty_rof(self):
        single = 'S' if self.rof[0] else '-'
        semi = str(self.rof[1]) if self.rof[1] else '-'
        auto = str(self.rof[2]) if self.rof[2] else '-'
        return f'{single}/{semi}/{auto}'

    @property
    def pretty_damage_type(self):
        return self.damage_type.name[0]


@require_kwargs
@dataclass(frozen=True)
class PlayerWeaponUpgrade(Item):
    pass


class PlayerWeaponInstance(InstanceMixin):

    def __init__(self, weapon_model: PlayerWeapon, *,
                       craftsmanship: Craftsmanship,
                       upgrades = None,
                       quantity: int = 1):

        self.weapon_model = weapon_model
        self.test_characteristic = Characteristic.WeaponSkill \
            if self.weapon_model.weapon_class == WeaponClass.Melee \
            else Characteristic.BallisticSkill

        self.upgrades = [] if upgrades is None else upgrades
        
        super().__init__(craftsmanship=craftsmanship,
                         quantity=quantity)

    def __str__(self):
        return f'<{self.name} {self.damage_roll}d10+{self.damage_bonus}{self.pretty_damage_type} {self.pretty_rof} {self.range}m>'

    @property
    def range(self):
        return self.weapon_model.weapon_range

    @property
    def name(self):
        return self.weapon_model.name

    @property
    def damage_bonus(self):
        return self.weapon_model.damage_bonus

    @property
    def damage_roll(self):
        return self.weapon_model.damage_roll

    @property
    def rof_single(self):
        return self.weapon_model.rof[0]

    @property
    def rof_semi(self):
        return self.weapon_model.rof[1]

    @property
    def pretty_rof(self):
        return self.weapon_model.pretty_rof

    @property
    def pretty_damage_type(self):
        return self.weapon_model.pretty_damage_type

    @property
    def rof_auto(self):
        return self.weapon_model.rof[2]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# zardoz/zardoz.py
# Copyright (c) 2021 Camille Scott <camille.scott.w@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import json
import os
from pathlib import Path
import sys
import textwrap

from discord.ext import commands

from . import __version__, __splash__, __about__, __testing__
from .rt.simulate import zimulate
from .utils import default_log_file, default_database_dir, EnumAction



class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def main():

    from .rt.combat import COMBAT_ACTIONS
    from .rt.items import ItemAvailability
    from .rt.weapons import (WeaponClass, WeaponType, DamageType,
                             Craftsmanship)

    parser = argparse.ArgumentParser(
        description=f'{__splash__}\n{__about__}',
        formatter_class=CustomFormatter
    )
    subparsers = parser.add_subparsers()

    bot = subparsers.add_parser('bot')
    bot.add_argument(
        '--secret-token',
        help='The secret token for the bot, '\
             'from the discord developer portal. '\
             'If you have set $ZARDOZ_TOKEN, this '\
             'option will override that.'
    )
    bot.add_argument(
        '--database-dir',
        type=lambda p: Path(p).absolute(),
        default=default_database_dir(debug=__testing__),
        help='Directory for the bot databases. Default follows the '\
             'XDG specifiction.'
    )
    bot.add_argument(
        '--log',
        type=lambda p: Path(p).absolute(),
        default=default_log_file(debug=__testing__),
        help='Path to the log file. Default follows the '\
             'XDG specifiction.'
    )
    bot.set_defaults(func=dize)

    simulate = subparsers.add_parser('zimulate')
    simulate.add_argument(
        '-BS',
        '--ballistic-skill',
        default=40,
        type=int
    )
    simulate.add_argument(
        '--actions',
        nargs='+',
        choices=COMBAT_ACTIONS.keys()
    )
    simulate.add_argument(
        '--target-range',
        default=10,
        type=int
    )
    simulate.add_argument(
        '--name',
        default='RT Weapon',
    )
    simulate.add_argument(
        '--availability',
        default=ItemAvailability.Scarce,
        action=EnumAction,
        type=ItemAvailability
    )
    simulate.add_argument(
        '--weapon-class',
        default=WeaponClass.Pistol,
        action=EnumAction,
        type=WeaponClass
    )
    simulate.add_argument(
        '--type',
        default=WeaponType.Las,
        type=WeaponType,
        action=EnumAction
    )
    simulate.add_argument(
        '--range',
        default=20,
        type=int
    )
    simulate.add_argument(
        '--rof-single',
        default=True,
        type=bool
    )
    simulate.add_argument(
        '--rof-semi',
        default=0,
        type=int
    )
    simulate.add_argument(
        '--rof-auto',
        default=0,
        type=int
    )
    simulate.add_argument(
        '--damage-d10',
        default=1,
        type=int
    )
    simulate.add_argument(
        '--damage-bonus',
        default=3,
        type=int
    )
    simulate.add_argument(
        '--damage-type',
        default=DamageType.Energy,
        type=DamageType,
        action=EnumAction
    )
    simulate.add_argument(
        '--pen',
        default=0,
        type=int
    )
    simulate.add_argument(
        '--clip',
        default=10,
        type=int
    )
    simulate.add_argument(
        '--reload-time',
        default=1.0,
        type=float
    )
    simulate.add_argument(
        '--mass',
        default=2.0,
        type=float
    )
    simulate.add_argument(
        '--craftsmanship',
        default=Craftsmanship.Common,
        type=Craftsmanship,
        action=EnumAction
    )
    simulate.add_argument(
        '--n-trials',
        '-N',
        default=10000,
        type=int
    )
    simulate.add_argument(
        '--plot',
        default='text',
        choices = ['text', 'image']
    )
    simulate.set_defaults(func=zimulate)

    args = parser.parse_args()
    args.func(args)


def dize(args):

    from .database import DatabaseCache
    from .logging import setup as setup_logger
    
    from .ztable import TableCommands
    from .zhistory import HistoryCommands
    from .zmode import ModeCommands
    from .zroll import RollCommands
    from .zvars import VarCommands

    log = setup_logger(args.log)

    TOKEN = args.secret_token
    if not TOKEN:
        try:
            TOKEN = os.environ['ZARDOZ_TOKEN']
        except KeyError:
            log.error('Must set ZARDOZ_TOKEN or use --secret-token')
            sys.exit(1)
        else:
            log.info('Got secret token from $ZARDOZ_TOKEN.')

    # get a handle for the history database
    DB = DatabaseCache(args.database_dir)

    class ZardozBot(commands.Bot):

        async def close(self):
            log.info('DatabaseCache close()')
            await DB.close()
            log.info('Bot close()')
            await super().close()
            log.info('Bot closed.')

    prefix = '/' if not __testing__ else '!'
    bot = ZardozBot(command_prefix=prefix)

    @bot.event
    async def on_ready():
        log.info(f'Ready: member of {bot.guilds}')
        log.info(f'Users: {bot.users}')

    @bot.command(name='zabout', help='Project info.')
    async def zabout(ctx):
        msg = f'version: {__version__}\n'\
              f'source: https://github.com/camillescott/zardoz/releases/tag/v{__version__}\n'\
              f'active installs: {len(bot.guilds)}'
        await ctx.message.reply(msg)

    bot.add_cog(RollCommands(bot, DB))
    bot.add_cog(TableCommands(bot, DB))
    bot.add_cog(VarCommands(bot, DB))
    bot.add_cog(ModeCommands(bot, DB))
    bot.add_cog(HistoryCommands(bot, DB))

    bot.run(TOKEN)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : logging.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import logging

from rich.logging import RichHandler


def setup(log_file):
    log_file.parent.mkdir(parents=True, exist_ok=True)

    logging.basicConfig(
        filename=log_file,
        level=logging.INFO,
        format='%(asctime)s %(levelname)10s [%(filename)s:%(lineno)s - %(funcName)20s()] %(message)s',
        datefmt="[%x %X]"
    )

    logger = logging.getLogger()
    logger.addHandler(RichHandler())

    return logger


class LoggingMixin:

    def __init__(self, *args, **kwargs):
        self.log = self.get_logger()
        self.register_decos()
        
    def register_decos(self):

        try:
            @self.bot.before_invoke
            async def log_cmd_invoke(ctx):
                self.log.info(f'CMD: [{ctx.invoked_with} {ctx.message.content}] from {ctx.author}:{ctx.guild}')
        except AttributeError:
            pass

    @staticmethod
    def get_logger():
        return logging.getLogger()



import os

__author__ = """Camille Scott"""
__email__ = 'cswel@ucdavis.edu'
__version__ = '0.8.1'

__splash__ = f'''
 _____             _      _____
|__  /__ _ _ __ __| | ___|__  /
  / // _` | '__/ _` |/ _ \ / / 
 / /| (_| | | | (_| | (_) / /_ 
/____\__,_|_|  \__,_|\___/____| v{__version__}
                                by {__author__}
'''
__about__ = 'A discord dice-rolling bot with lots of nifty features.'
__testing__ = bool(os.environ.get('ZARDOZ_DEBUG', False))

from .zardoz import main



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2021
# File   : character.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 25.02.2021

from dataclasses import dataclass
from enum import Enum

from ..utils import require_kwargs


class Characteristic(Enum):
    WeaponSkill = 'WS'
    BallisticSkill = 'BS'
    Strength = 'S'
    Toughness = 'T'
    Agility = 'AG'
    Intelligence = 'INT'
    Perception = 'PER'
    Willpower = 'WP'
    Fellowship = 'FEL'


class CharacteristicInstance:

    def __init__(self, characteristic: Characteristic, base: int):
        self.characteristic = characteristic
        self.base = base
        self.modifiers = []
        self.modifier_tags = []
        self.advances = 0
        self.multiplier = 1

    def __int__(self):
        return self.base + (self.advances * 5) + sum(self.modifiers)

    @property
    def bonus(self):
        return (int(self) // 10) * self.multiplier

    def advance(self):
        if self.advances < 4:
            self.advances += 1

    def add_modifier(self, modifier: int, tag: str = ''):
        self.modifiers.append(modifier)
        self.modifier_tags.append(tag)

    def clear_modifiers(self):
        self.modifiers = []
        self.modifier_tags = []

    def __str__(self):
        return f'{self.characteristic.value}={int(self)}: '\
               f'bonus={self.bonus}, advances={self.advances}, '\
               f'modifiers={dict(zip(self.modifier_tags, self.modifiers))}'


class Player:

    def __init__(self, *,
                 name: str,
                 base_ws: int,
                 base_bs: int,
                 base_s: int,
                 base_t: int,
                 base_ag: int,
                 base_int: int,
                 base_per: int,
                 base_wp: int,
                 base_fel: int,
                 base_fate: int):

        if name == '':
            raise ValueError('Player must have a name!')

        self.name = name
        self.chars = {
            Characteristic.WeaponSkill.value: CharacteristicInstance(Characteristic.WeaponSkill, base_ws),
            Characteristic.BallisticSkill.value: CharacteristicInstance(Characteristic.BallisticSkill, base_bs),
            Characteristic.Strength.value: CharacteristicInstance(Characteristic.Strength, base_s),
            Characteristic.Toughness.value: CharacteristicInstance(Characteristic.Toughness, base_t),
            Characteristic.Agility.value: CharacteristicInstance(Characteristic.Agility, base_ag),
            Characteristic.Intelligence.value: CharacteristicInstance(Characteristic.Intelligence, base_int),
            Characteristic.Perception.value: CharacteristicInstance(Characteristic.Perception, base_per),
            Characteristic.Willpower.value: CharacteristicInstance(Characteristic.Willpower, base_wp),
            Characteristic.Fellowship.value: CharacteristicInstance(Characteristic.Fellowship, base_fel)
        }
        for name, characteristic in self.chars.items():
            setattr(self, name, characteristic)
        
        self.base_fate = base_fate if base_fate > 0 else 0

    def __str__(self):
        chars = '\n    '.join((str(char) for char in self.chars.values()))
        return f'Player: {self.name}\n'\
                '  Characteristics:\n'\
               f'    {chars}\n'\
               f'  Fate: {self.base_fate}'



# SPDX-License-Identifier: CC0-1.0
# Source: https://github.com/jktr/matplotlib-backend-kitty

from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import (
    _Backend, FigureCanvasBase, FigureManagerBase)
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib import interactive

from io import BytesIO
from subprocess import run

import sys


# XXX heuristic for interactive repl
if sys.flags.interactive:
    interactive(True)


class FigureManagerICat(FigureManagerBase):

    @classmethod
    def _run(self, *cmd):
        def f(*args, output=True, **kwargs):
            if output:
                kwargs['capture_output'] = True
                kwargs['text'] = True
            r = run(cmd + args, **kwargs)
            if output:
                return r.stdout.rstrip()
        return f

    def show(self):
        tput = __class__._run('tput')
        icat = __class__._run('kitty', '+kitten', 'icat')

        # gather terminal dimensions
        rows = int(tput('lines'))
        px = icat('--print-window-size')
        px = list(map(int, px.split('x')))

        # account for post-display prompt scrolling
        # 3 line shift for [\n, <matplotlib.axes…, >>>] after the figure
        px[1] -= int(3*(px[1]/rows))

        # resize figure to terminal size & aspect ratio
        dpi = self.canvas.figure.dpi
        self.canvas.figure.set_size_inches(((px[0] / dpi) * .75, (px[1] / dpi) * .9))

        with BytesIO() as buf:
            self.canvas.figure.savefig(buf, format='png', transparent=True, facecolor='None')
            icat('--align', 'left', output=False, input=buf.getbuffer())

@_Backend.export
class _BackendICatAgg(_Backend):
    FigureCanvas = FigureCanvasAgg
    FigureManager = FigureManagerICat

    # XXX: `trigger_manager_draw` is intended
    # for updates, not one-shot rendering.
    # Thus, we need to filter calls for figures
    # that aren't fully initialized yet, like the
    # call from `plt.figure`. Our heuristic for
    # an initialized figure is the presence of axes.
    def _trigger_manager_draw(manager):
        if manager.canvas.figure.get_axes():
            manager.show()
            Gcf.destroy(manager)

    def show(*args, **kwargs):
        _Backend.show(*args, **kwargs)
        Gcf.destroy_all()

    trigger_manager_draw = _trigger_manager_draw
    mainloop = lambda: None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2021
# File   : items.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 19.03.2021

from dataclasses import dataclass, field
from enum import IntEnum

from ..utils import require_kwargs


class Craftsmanship(IntEnum):
    Poor = 10
    Common = 0
    Good = -10
    Best = -30


class ItemAvailability(IntEnum):
    Ubiquitous = 70
    Abundant = 50
    Plentiful = 30
    Common = 20
    Average = 10
    Scarce = 0
    Rare = -10
    VeryRare = -20
    ExtremelyRare = -30
    NearUnique = -50
    Unique = -70


class ItemScale(IntEnum):
    Negligible = 30
    Trivial = 20
    Minor = 10
    Standard = 0
    Major = - 10
    Signficant = -20
    Vast = -30

    @staticmethod
    def from_quantity(q):
        if q == 1:
            return ItemScale.Negligible
        elif 2 <= q < 9:
            return ItemScale.Trivial
        elif 10 <= q < 30:
            return ItemScale.Minor
        elif 30 <= q < 500:
            return ItemScale.Standard
        elif 500 <= q < 2000:
            return ItemScale.Major
        elif 2000 <= q < 10000:
            return ItemScale.Signficant
        else:
            return ItemScale.Vast


@require_kwargs
@dataclass(frozen=True)
class Item:

    name: str
    availability: ItemAvailability
    mass: float

    def roll_acquisition(self,
                         quantity: int,
                         profit_factor: int,
                         craftsmanship: Craftsmanship = Craftsmanship.Common,
                         modifier: int = 0):

        roll = dice.roll('1d100')
        return int(roll) <= profit_factor + quantity + self.availability + craftsmanship + modifier


class InstanceMixin:

    def __init__(self, *,
                 craftsmanship: Craftsmanship,
                 quantity: int = 1,
                 **kwargs):
        
        self.craftsmanship = craftsmanship
        self.quantity = quantity
        
        super().__init__(**kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : zcrit.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

from collections import OrderedDict
import glob
import os
import typing

import dice
import yaml

from discord.ext import commands

from .logging import LoggingMixin
from .rolls import SimpleRollConvert
from .utils import __pkg_dir__


class RollTable:

    def __init__(self, table_yaml):
        with open(table_yaml) as fp:
            data = yaml.safe_load(fp)

        self.full_name = data['full_name']
        self.slug = data['slug']
        self.game = data['game']
        self.book = data['book']
        self.die = data['die']
        self.rolls = data['rolls']
    
    
    def get(self, roll):
        for option in self.rolls:
            lower, upper = option['range']
            name = '' if not option['name'] else option['name'].strip()
            if lower <= roll <= upper:
                return name, ' '.join(option['effect'].split())

        raise ValueError(f'No entry for {roll}')

    def roll(self, modifers = None):
        rolled_val = int(dice.roll(self.die))
        name, effect = self.get(rolled_val)

        return rolled_val, name, effect


def load_crit_tables(log=None):
    
    files = glob.glob(os.path.join(__pkg_dir__, 'tables', '*.yaml'))
    tables = {}
    for file_path in files:
        table = RollTable(file_path)
        tables[table.slug] = table
    return OrderedDict(sorted(tables.items(), key=lambda tup: tup[0]))


TABLES = load_crit_tables()


class TableConvert(commands.Converter):

    async def convert(self, ctx, argument):
        try:
            table = TABLES[argument]
        except KeyError:
            raise commands.BadArgument(f'{argument} is not a valid table.')
        return table


class TableCommands(commands.Cog, LoggingMixin):

    def __init__(self, bot, db):
        self.bot = bot
        self.db = db
        super().__init__()

        self.log.info(f'Loaded tables: {TABLES}')

    @commands.command(name='ztable', help='Roll on a table.')
    async def ztable(self, ctx, table: TableConvert = None,
                               val: typing.Union[int, SimpleRollConvert] = None):
        if ctx.invoked_subcommand is not None:
            return

        if table is None:
            header = '**Available Tables:**'
            body = []
            for slug, table in TABLES.items():
                body.append(f'`{slug:15}` {table.full_name} ({table.game}, {table.book})')
            body = '\n'.join(body)
            await ctx.message.reply(f'{header}\n{body}')

            self.log.info(f'/ztable: {table.slug}, {val}')
            return

        try:
            if val is None:
                val, name, effect = table.roll()
                result = f'{table.die} ⤳ {val}'
            else:
                name, effect = table.get(int(val))
                result = f'{val}'
        except ValueError:
            await ctx.message.reply(f'Bad table value. Perils be upon ye.')
        else:
            msg = [f'**Roll:** {result}',
                   f'**Table:** {table.full_name} ({table.game}, {table.book})']
            if name:
                msg.append(f'**Name:** {name}')
            msg.append(f'**Effect:** {effect}')

            await ctx.message.reply('\n'.join(msg))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2021
# File   : combat.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 25.02.2021

from abc import ABC, abstractmethod
import collections
import os

from .character import Characteristic
from ..utils import __pkg_dir__, reverse_number, d10, d100, Nd10
from ..ztable import RollTable


class CharacteristicBonus:

    def __init__(self, characteristic: Characteristic, bonus: int):
        self.characteristic = characteristic
        self.bonus = bonus

    def __int__(self):
        return self.bonus

    def __call__(self, ctx, **kwargs):
        if self.characteristic == ctx.characteristic:
            ctx.add_test_bonus(self.bonus)

    def __str__(self):
        return f'Bonus: {self.characteristic.name} {self.bonus}'


class ExtraHitsBonus:

    def __init__(self, dos_div: int = 1):
        self.dos_div = 1

    def __call__(self, ctx, **kwargs):
        ctx.add_hits(ctx.attack_degrees  // self.dos_div)

    def __str__(self):
        return f'Bonus: 1 hit per {self.dos_div} DoS'


COMBAT_ACTIONS = {}

class CombatAction:

    def __init__(self, name: str, before_effects = None, after_effects = None, special=''):
        self.name = name
        self.before_effects = before_effects if before_effects else []
        self.after_effects = after_effects if after_effects else []
        self.special = special

        COMBAT_ACTIONS[name] = self

    def effects(self):
        if self.before_effects:
            for effect in self.before_effects:
                yield effect
        if self.after_effects:
            for effect in self.after_effects:
                yield effect


AimFull = CombatAction('Aim Full',
                       before_effects = [
                           CharacteristicBonus(Characteristic.WeaponSkill, 20),
                           CharacteristicBonus(Characteristic.BallisticSkill, 20)
                       ])

AimHalf = CombatAction('Aim Half',
                       before_effects = [
                           CharacteristicBonus(Characteristic.WeaponSkill, 10),
                           CharacteristicBonus(Characteristic.BallisticSkill, 10)
                       ])

AllOutAttack = CombatAction('All Out Attack',
                            before_effects = [
                                CharacteristicBonus(Characteristic.WeaponSkill, 20)
                            ],
                            special='Cannot dodge or parry')

CalledShot = CombatAction('Called Shot',
                          before_effects = [
                              CharacteristicBonus(Characteristic.WeaponSkill, 20),
                              CharacteristicBonus(Characteristic.BallisticSkill, 20)
                          ],
                          special='Attack a specific location')

Charge = CombatAction('Charge',
                      before_effects = [
                          CharacteristicBonus(Characteristic.WeaponSkill, 10)
                      ],
                      special='Must move 4 metres')

FullAutoBurst = CombatAction('Full Auto Burst',
                             before_effects  = [
                                 CharacteristicBonus(Characteristic.BallisticSkill, 20)
                             ],
                             after_effects = [
                                 ExtraHitsBonus()
                             ])

SemiAutoBurst = CombatAction('Semi Auto Burst',
                             before_effects = [
                                 CharacteristicBonus(Characteristic.BallisticSkill, 10)
                             ],
                             after_effects = [
                                 ExtraHitsBonus(dos_div=2)
                             ])


class HitLocTable:

    def __init__(self):

        self.base = RollTable(os.path.join(__pkg_dir__, 'tables', 'rt_hit_loc.yaml'))
        self.table = {'Head': ['Head', 'Arm', 'Body', 'Arm', 'Body'],
                      'Arm':  ['Arm', 'Body', 'Head', 'Body', 'Arm'],
                      'Body': ['Body', 'Arm', 'Head', 'Arm', 'Body'],
                      'Leg':  ['Leg', 'Body', 'Arm', 'Head', 'Body']}

    def get_location(self, init_roll, n_hits):
        table_val = reverse_number(init_roll)
        locs = []
        _, init_loc = self.base.get(table_val)
        locs.append(init_loc)

        if n_hits > 1:
            x_hits = n_hits - 1
            idx_loc = init_loc
            if 'Arm' in idx_loc:
                idx_loc = 'Arm'
            if 'Leg' in idx_loc:
                idx_loc = 'Leg'

            subtable = self.table[idx_loc]
            locs.extend(subtable[:x_hits])
            if x_hits > len(subtable):
                locs.extend([subtable[-1]] * (x_hits - len(subtable)))

        return locs
    
HIT_LOC_TABLE = HitLocTable()


class DamageRoll:

    def __init__(self, dice, bonus):
        self.bonus = bonus
        self.fury_bonus = 0
        self.dice = dice

    def replace_lowest(self, replacement):
        if min(self.dice) < replacement:
            self.dice[self.dice.index(min(self.dice))] = replacement

    def __int__(self):
        return sum(self.dice) + self.bonus + self.fury_bonus

    def add_fury(self, extra_fury):
        self.fury_bonus += extra_fury

    def add_bonus(self, extra_bonus):
        self.bonus += extra_bonus


class AttackContext:

    def __init__(self, weapon, test_base, target_range, actions = [], quiet: bool = True):
        self.characteristic = weapon.test_characteristic
        self.test_base = test_base
        self.weapon = weapon
        self._test_bonus = 0
        self.target_range = target_range
        self.hits_base = 0
        self.hits_extra = 0
        self.actions = actions

        self.attack_roll = 0
        self.attack_degrees = 0

        self.damage = 0
        self.damage_bonus = 0
        self.damage_rolls = []

    def add_test_bonus(self, extra):
        self._test_bonus += extra

    def add_hits(self, extra):
        self.hits_extra += extra

    def add_damage_roll(self, roll):
        self.damage_rolls.append(roll)

    @property
    def test_bonus(self):
        return min(60, self._test_bonus)

    @property
    def test(self):
        return self.test_base + self.test_bonus

    @property
    def hits(self):
        if SemiAutoBurst in self.actions:
            hits_max = self.weapon.rof_semi
        elif FullAutoBurst in self.actions:
            hits_max = self.weapon.rof_auto
        else:
            hits_max = int(self.weapon.rof_single)
        return min(self.hits_base + self.hits_extra, hits_max)

    @property
    def success(self):
        return self.attack_roll <= self.test

    @property
    def total_damage(self):
        return sum((int(roll) for roll in self.damage_rolls))


def player_attack(weapon_instance, char_val, actions=None, target_range: int = 10, quiet: bool = True):

        def _print(*args, **kwargs):
            if not quiet:
                print(*args, **kwargs)

        #
        # Sanitize parameters
        #

        if actions is None:
            actions = []
        elif not isinstance(actions, collections.Sequence):
            actions = set([actions])
        else:
            actions = set(actions)

        assert len(actions) < 3

        if SemiAutoBurst in actions and not weapon_instance.rof_semi:
            raise ValueError(f'{weapon_instance.name} does not support semi-auto')
        if FullAutoBurst in actions and not weapon_instance.rof_auto:
            raise ValueError(f'{weapon_instance.name} does not support full-auto')
        if target_range / weapon_instance.range > 4:
            raise ValueError(f'{weapon_instance.name} cannot fire more than {weapon_instance.range * 4}m')

        # setup the attack context
        ctx = AttackContext(weapon_instance, char_val, target_range, actions=actions, quiet=quiet)

        # apply characteristic bonuses
        for action in actions:
            for bonus in action.before_effects:
                bonus(ctx)
                _print(bonus)
        # now we'd apply specials from the weapon itself in the same way

        # figure ranges
        if target_range <= 2:
            # point blank
            _print('Point blank: +30')
            ctx.add_test_bonus(30)
        elif target_range <= (weapon_instance.range / 2):
            # short range
            ctx.add_test_bonus(10)
            _print('Close: +10')
        elif target_range >= (weapon_instance.range * 3):
            # extreme range
            ctx.add_test_bonus(-30)
            _print('Extreme: -30')
        elif target_range >= (weapon_instance.range * 2):
            # long range
            ctx.add_test_bonus(-10)
            _print('Long: -10')
        else:
            _print('Normal range.')

        _print(f'final test: {ctx.test_base} + {ctx.test_bonus}')

        # roll it
        ctx.attack_roll = d100()
        delta = abs(ctx.test - ctx.attack_roll)
        ctx.attack_degrees = delta // 10

        _print(f'rolled {ctx.attack_roll} for {ctx.attack_degrees} degrees')

        # TODO: check jamming and exploding

        if not ctx.success:
            # damage, effective_char, degrees, hits, locations, message
            return 0, ctx.test, ctx.attack_degrees, 0, [], "Failed to hit"
        else:
            ctx.hits_base = 1

        # apply action bonuses
        for action in actions:
            for bonus in action.after_effects:
                bonus(ctx)
                _print(bonus)

        _print(f'hits after bonus: {ctx.hits}')

        # get hit location with reverse_number
        # for multiple hits 

        # roll for damage
        for hit_counter in range(ctx.hits):
            _print(f'roll hit {hit_counter + 1}')

            roll = DamageRoll(Nd10(n=weapon_instance.damage_roll), weapon_instance.damage_bonus)

            _print(f'initial damage roll: {roll.dice}')

            # replace lowest with DoS if its lower
            roll.replace_lowest(ctx.attack_degrees)

            _print(f'damage roll after replace: {roll.dice}')

            # TODO: apply weapon special roll bonuses

            ctx.add_damage_roll(roll)
            _print(f'damage before fury: {int(roll)}')

            fury = 10 in roll.dice
            while (fury):
                fury_roll = d100() <= ctx.test

                if fury_roll:
                    fury_damage = d10()

                    _print(f'fury! rolled {fury_damage}')

                    roll.add_fury(fury_damage)
                    fury = fury_damage == 10
                else:
                    fury = False

        _print(f'final damage: {ctx.total_damage}')
                

        locs = HIT_LOC_TABLE.get_location(ctx.attack_roll, ctx.hits)
        return ctx.total_damage, ctx.test, ctx.attack_degrees, ctx.hits, locs, "Hit!"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2020
# File   : utils.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 22.02.2021

import argparse
from dataclasses import is_dataclass
from enum import Enum
from typing import TypeVar, Type, Callable, List, Dict, Any
from datetime import datetime
import functools
import os
from pathlib import Path
import random

import discord
from xdg import xdg_data_home


__pkg_dir__ = os.path.abspath(os.path.dirname(__file__))
__time_format__ = '%a %b %d %I:%M%p %Z'

_T = TypeVar("_T")
_Self = TypeVar("_Self")
_VarArgs = List[Any]
_KWArgs = Dict[str, Any]


FAILURE = 'F 🛇'
SUCCESS = 'S ✅'


def d10():
    return random.randint(1, 10)


def d100():
    return random.randint(1, 100)


def Nd10(n=1):
    return [random.randint(1,10) for _ in range(n)]


def Nd100(n=1):
    return [random.randint(1,100)  for _ in range(n)]


def default_database_dir(debug=False):
    if not debug:
        return xdg_data_home().joinpath('zardoz-bot', 'databases')
    else:
        return xdg_data_home().joinpath('zardoz-bot', 'debug', 'databases')


def default_log_file(debug=False):
    if not debug:
        return xdg_data_home().joinpath('zardoz-bot', 'bot.log')
    else:
        return xdg_data_home().joinpath('zardoz-bot', 'debug', 'bot.log')


def reverse_number(num: int):
    result = 0
    while (num > 0):
        result = (num * 10) + (num % 10)
        num = num // 10
    return result


def handle_http_exception(func):

    @functools.wraps(func)
    async def wrapper(self, ctx, *args, **kwargs):
        try:
            return await func(self, ctx, *args, **kwargs)
        except discord.HTTPException:
            await ctx.message.reply('Hey broh, that\'s a big response.'\
                                    ' Try dialing back your request,'\
                                    ' I\'m only human.')
            raise
    return wrapper


class EnumAction(argparse.Action):
    """
    Argparse action for handling Enums
    """
    def __init__(self, **kwargs):
        # Pop off the type value
        enum = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum is None:
            raise ValueError("type must be assigned an Enum when using EnumAction")
        if not issubclass(enum, Enum):
            raise TypeError("type must be an Enum when using EnumAction")

        # Generate choices from the Enum
        kwargs.setdefault("choices", tuple(e.name for e in enum))

        super(EnumAction, self).__init__(**kwargs)

        self._enum = enum

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        enum = self._enum[values]
        setattr(namespace, self.dest, enum)


#
# Dataclass utilities taken from: https://gist.github.com/mikeholler/4be180627d3f8fceb55704b729464adb
#

def _kwarg_only_init_wrapper(
        self: _Self,
        init: Callable[..., None],
        *args: _VarArgs,
        **kwargs: _KWArgs
) -> None:
    if len(args) > 0:
        raise TypeError(
            f"{type(self).__name__}.__init__(self, ...) only allows keyword arguments. Found the "
            f"following positional arguments: {args}"
        )
    init(self, **kwargs)


def _positional_arg_only_init_wrapper(
        self: _Self,
        init: Callable[..., None],
        *args: _VarArgs,
        **kwargs: _KWArgs
) -> None:
    if len(kwargs) > 0:
        raise TypeError(
            f"{type(self).__name__}.__init__(self, ...) only allows positional arguments. Found "
            f"the following keyword arguments: {kwargs}"
        )
    init(self, *args)


def require_kwargs(cls: Type[_T]) -> Type[_T]:
    """
    Force a dataclass's init function to only work if called with keyword arguments.
    If parameters are not positional-only, a TypeError is thrown with a helpful message.
    This function may only be used on dataclasses.
    This works by wrapping the __init__ function and dynamically replacing it. Therefore,
    stacktraces for calls to the new __init__ might look a bit strange. Fear not though,
    all is well.
    Note: although this may be used as a decorator, this is not advised as IDEs will no longer
    suggest parameters in the constructor. Instead, this is the recommended usage::
        from dataclasses import dataclass
        @dataclass
        class Foo:
            bar: str
        require_kwargs_on_init(Foo)
    """

    if cls is None:
        raise TypeError("Cannot call with cls=None")
    if not is_dataclass(cls):
        raise TypeError(
            f"This decorator only works on dataclasses. {cls.__name__} is not a dataclass."
        )

    original_init = cls.__init__

    def new_init(self: _Self, *args: _VarArgs, **kwargs: _KWArgs) -> None:
        _kwarg_only_init_wrapper(self, original_init, *args, **kwargs)

    # noinspection PyTypeHints
    cls.__init__ = new_init  # type: ignore

    return cls


def require_positional_args(cls: Type[_T]) -> Type[_T]:
    """
    Force a dataclass's init function to only work if called with positional arguments.
    If parameters are not positional-only, a TypeError is thrown with a helpful message.
    This function may only be used on dataclasses.
    This works by wrapping the __init__ function and dynamically replacing it. Therefore,
    stacktraces for calls to the new __init__ might look a bit strange. Fear not though,
    all is well.
    Note: although this may be used as a decorator, this is not advised as IDEs will no longer
    suggest parameters in the constructor. Instead, this is the recommended usage::
        from dataclasses import dataclass
        @dataclass
        class Foo:
            bar: str
        require_positional_args_on_init(Foo)
    """

    if cls is None:
        raise TypeError("Cannot call with cls=None")
    if not is_dataclass(cls):
        raise TypeError(
            f"This decorator only works on dataclasses. {cls.__name__} is not a dataclass."
        )

    original_init = cls.__init__

    def new_init(self: _Self, *args: _VarArgs, **kwargs: _KWArgs) -> None:
        _positional_arg_only_init_wrapper(self, original_init, *args, **kwargs)

    # noinspection PyTypeHints
    cls.__init__ = new_init  # type: ignore

    return cls



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Camille Scott, 2021
# File   : simulate.py
# License: MIT
# Author : Camille Scott <camille.scott.w@gmail.com>
# Date   : 27.02.2021

from contextlib import contextmanager
import os


from .combat import player_attack, COMBAT_ACTIONS
from .items import ItemAvailability
from .weapons import (WeaponClass, WeaponType, DamageType, Craftsmanship,
                      PlayerWeapon, PlayerWeaponInstance)

def zimulate(args):

    weapon_model = PlayerWeapon(
        name=args.name,
        availability=args.availability,
        weapon_class=args.weapon_class,
        weapon_type=args.type,
        weapon_range=args.range,
        rof=(args.rof_single, args.rof_semi, args.rof_auto),
        damage_roll=args.damage_d10,
        damage_bonus=args.damage_bonus,
        damage_type=args.damage_type,
        pen=args.pen,
        clip=args.clip,
        reload_time=args.reload_time,
        mass=args.mass
    )

    weapon_instance = PlayerWeaponInstance(weapon_model, craftsmanship=args.craftsmanship)
    combat_actions = [COMBAT_ACTIONS.get(action) for action in args.actions] \
                     if args.actions is not None else []

    results = simulate_attack(weapon_instance, args.ballistic_skill,
                              args.target_range, combat_actions, N=args.n_trials)
    sr = sum(results.damage > 0) / args.n_trials
    maxd = results.damage.max()
    medd = int(results.damage[results['damage'] > 0].median())

    actions_str = ', '.join([action.name for action in combat_actions])
    title = f'{weapon_instance.name} @ {actions_str}: SR={sr:.3f}, Max={maxd}, Med={medd}\n'\
            f'BS={args.ballistic_skill}, Range={args.target_range}, N={args.n_trials:,} '

    if args.plot is not None:
        if args.plot == 'text':
            plot_simulation_plottile(results)
        else:
            plot_simulation_mpl(results, title=title)


def simulate_attack(instance, BS, target_range, actions, N=10000):
    from alive_progress import alive_bar
    import pandas as pd

    damages, tests = [], []
    with alive_bar(N, title=f'{instance}') as bar:
        for _ in range(N):
            dmg, test, _, _, _, _ = player_attack(instance, BS, actions=actions, target_range=target_range)
            damages.append(dmg)
            tests.append(test)
            bar()
    return pd.DataFrame({'damage': damages, 'test': tests})


def plot_simulation_mpl(results_df, title='', filename=None, **kwargs):
    import matplotlib as mpl
    if os.environ['TERM'] == 'xterm-kitty':
        mpl.use('module://zardoz.mpl-kitty')
    else:
        mpl.use('module://zardoz.mpl-sixel')
    import matplotlib.pyplot as plt
    plt.ioff()
    #import ficus
    import seaborn as sns

    #filename=f'BS{BS}_{weapon_instance.name}_R{target_range}_.png'
    sns.set_style('ticks')
    sns.set_context('talk')

    with mpl_dark_mode():
        fig, ax = plt.subplots(figsize=(12,8))
        sns.kdeplot(data=results_df[results_df['damage'] != 0], x='damage', fill=True, ax=ax)
        sns.despine(ax=ax, offset=10)
        ax.set_xlabel('Damage')
        ax.set_title(title)
        plt.show()


def plot_simulation_plottile(results_df, title='', **kwargs):
    import plotille
    print(title)
    print(plotille.histogram(results_df.damage[results_df.damage > 0],
                             X_label='Damage',
                             x_min=0,
                             height=30))


@contextmanager
def mpl_dark_mode(*args, **kwargs):
    import matplotlib.pyplot as plt
    line_color = 'white'
    with plt.rc_context({'font.family': 'monospace', 'text.color': line_color, 'axes.labelcolor': line_color,
                         'xtick.color': line_color, 'ytick.color': line_color,
                         'axes.edgecolor': line_color, 'axes.facecolor': '#000000',
                         'figure.facecolor': '#000000'}):
        yield line_color
